/*
 * Created on Jan 28, 2005
 *
 * (C) 2005 Amin Ahmad and Ahmadsoft Partners Ltd.
 */
package org.ahmadsoft.foprocessor.ui.dialogs;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.concurrent.CountDownLatch;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;

import org.ahmadsoft.foprocessor.FoProcessorPlugin;
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.debug.internal.ui.preferences.IDebugPreferenceConstants;
import org.eclipse.debug.ui.IDebugUIConstants;
import org.eclipse.debug.ui.console.ConsoleColorProvider;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.operation.ModalContext;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.swt.widgets.Shell;

/**
 *
 * @author Amin Ahmad
 */
public class RenderProgress extends TitleAreaDialog implements IRunnableContext {

    private Image image = null;
    private ITextViewer console;
    private Document document;
    
    private ProgressBar progressBar;
    private Label consoleTitle;
    
    private ConsoleColorProvider colorProvider = new ConsoleColorProvider();
    
    private CountDownLatch createSignal = new CountDownLatch(1);
    
    /**
     * @param parentShell
     */
    public RenderProgress(Shell parentShell) {
        super(parentShell);        
        setShellStyle(SWT.TITLE | SWT.BORDER | SWT.APPLICATION_MODAL | SWT.RESIZE);
    }
    
    protected Point getInitialSize() {
        return getShell().computeSize(460, 500, true);
    }
    
    protected Control createContents(Composite parent) {
        Control result = super.createContents(parent);
        
        createSignal.countDown();
        
        return result;
    }
    
    protected Control createDialogArea(Composite parent) {
        Composite compositeParent = (Composite)super.createDialogArea(parent);
        
        // Setup the title area
        //
        setTitleImage(getImage());
        setTitle("Rendering"); 
        setMessage("Rendering documents."); 
        
        Composite pageContainer = new Composite(compositeParent, SWT.NONE);
        GridData gd = new GridData(GridData.FILL_BOTH);
        pageContainer.setLayoutData(gd);
        pageContainer.setFont(parent.getFont());       
        
        // Document 
        //
        document = new Document();
        
        //layout for page
        //
        GridLayout layout = new GridLayout();
        layout.numColumns = 1;
        
        pageContainer.setLayout(layout);
        
        progressCaption = new Label(pageContainer, SWT.NONE);
        progressCaption.setText("Progress:");
        progressCaption.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        
        progressBar = new ProgressBar(pageContainer, SWT.SMOOTH | SWT.HORIZONTAL);
        gd = new GridData(GridData.FILL_HORIZONTAL);
        gd.heightHint = 10;
        progressBar.setLayoutData(gd);
        
        consoleTitle = new Label(pageContainer, SWT.NONE);
        consoleTitle.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        
        console = new TextViewer(pageContainer, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
        console.getTextWidget().setFont(JFaceResources.getFont(IDebugPreferenceConstants.CONSOLE_FONT));
        console.getTextWidget().setLayoutData(new GridData(GridData.FILL_BOTH));
        console.setDocument(document);
         
        // stream.setColor(fColorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
        
        
        // Build the separator line
        //
        Label separator= new Label(compositeParent, SWT.HORIZONTAL | SWT.SEPARATOR);
        separator.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        
        return compositeParent;
    }

    public Image getImage() {
        if (image == null) {
            image = FoProcessorPlugin.getDefault().getImageDescriptor("full/wizban/RenderProgress.gif").createImage();
        }
        return image;
    }
    
    public void setConsoleTitle() {
        SimpleDateFormat df = new SimpleDateFormat("MMM dd, yyyy HH:mm:ss");
        consoleTitle.setText("XSL-FO Rendering [Powered by Apache FOP] (" +df.format(new Date())+ ")");
    }
    
    /**
     * Renders the document. Should not be run on the UI thread.
     * 
     * @param mimeType
     * @param extension
     * @param resources
     */
    public void beginRendering(final String mimeType, final String extension, final IStructuredSelection resources) {
        try {
            createSignal.await();
        } catch (InterruptedException e) {
            error(null, e);
        }
        
        // Set the titles and the button states
        //
        setConsoleTitle();
        
        Button okButton = getButton(IDialogConstants.OK_ID);
        if (okButton != null) {
            okButton.setEnabled(false);
        }
        
        Button cnButton = getButton(IDialogConstants.CANCEL_ID);
        if (cnButton != null) {
            cnButton.setEnabled(true);
        }
        
        IRunnableWithProgress runnable = new IRunnableWithProgress() {
            public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
                
                monitor.beginTask("Initializing", 3);
                
                // Setup logger. This is a hack because
                // per-instance logging is not supported.
                //
                Handler lh = new LogHandler();
                Logger.getLogger("org.apache.fop").addHandler(lh);
                
                monitor.worked(1);
                
                // Determine information about the files.
                //
                long totalByteSize = 0;
                for (Iterator i=resources.iterator(); i.hasNext(); ) {
                    IResource resource = (IResource) i.next();
                    if (resource instanceof IFile) {
                        IFile inputFile  = (IFile) resource;
                        File file = inputFile.getRawLocation().toFile();
                        
                        if (file.exists()) {
                            totalByteSize += file.length();
                        }
                    }
                }
                
                monitor.worked(2);

                monitor.beginTask("Rendering", totalByteSize > Integer.MAX_VALUE ? Integer.MAX_VALUE: (int) totalByteSize);
                
                // Rendition loop
                //
                for (Iterator i=resources.iterator(); i.hasNext(); ) {
                    IResource resource = (IResource) i.next();
                    if (resource instanceof IFile) {                
                        IFile inputFile  = (IFile) resource;
                        IPath inputPath  = inputFile.getRawLocation();
                        IPath outputPath = inputPath.removeFileExtension().addFileExtension(extension);

                        // Setup my driver.
                        //
                        FOUserAgent ua = new FOUserAgent();
                        ua.setOutputFile(outputPath.toFile());
                        
                        Fop driver = new Fop(mimeType,ua);
                        
                        //IFile outputFile = resource.getProject().getFile(outputPath);
                        
                        OutputStream out = null;
                        InputStream  in  = null;
                        
                        File file = outputPath.toFile();
                        try {
                            out = new BufferedOutputStream(new FileOutputStream(file));
                            driver.setOutputStream(out);

                            // Setup JAXP using identity transformer
                            TransformerFactory factory = TransformerFactory.newInstance();
                            Transformer transformer = factory.newTransformer(); // identity transformer
                            
                            // Setup input stream
                            Source src = new StreamSource(inputFile.getContents());

                            // Resulting SAX events (the generated FO) must be piped through to FOP
                            Result res = new SAXResult(driver.getDefaultHandler());
                            
                            // Start XSLT transformation and FOP processing
                            monitor.subTask(outputPath.lastSegment());
                            transformer.transform(src, res);
                            
                            monitor.worked((int)inputFile.getRawLocation().toFile().length());
                        } catch (Exception e) {
                            error(null, e);
                            continue;
                        } finally {
                            try {
                                if (out != null) out.close();
                                if (in != null) in.close();
                            } catch (IOException e) {
                                error(null, e);
                            }
                            
                            try {
                                inputFile.getParent().refreshLocal(1, new NullProgressMonitor());
                            } catch (CoreException e) {
                                error(null, e);
                            }
                        }
                    }
                }
                
                monitor.done();
                
                Display.getDefault().asyncExec(
                    new Runnable() {
                        public void run() {
                            Button okButton = getButton(IDialogConstants.OK_ID);
                            if (okButton != null) {
                                okButton.setEnabled(true);
                                okButton.setFocus();
                            }
                            
                            Button cnButton = getButton(IDialogConstants.CANCEL_ID);
                            if (cnButton != null) {
                                cnButton.setEnabled(false);
                            }
                        }
                    }
                );
            }
        };
        
        try {
            run(true, true, runnable);
        } catch (Exception e) {
            error(null, e);
        }
    }

    
    /**
     * @inheritDoc
     */
    public boolean close() {
        getImage().dispose();
        return super.close();
    }

    // IRunnableContext
    //
    
    public void run(boolean fork, boolean cancelable, IRunnableWithProgress runnable) throws InvocationTargetException, InterruptedException {
        // Create the Progress Monitor
        //
        ProgressMonitor monitor = new ProgressMonitor();
        
        ModalContext.run(runnable, fork, monitor, getShell().getDisplay());
    }
    
    // Logging Methods --------------------------------------------------------

    private static final String nl = System.getProperty("line.separator");
    private Label progressCaption;
    
    private String toString(Throwable t) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(bos);
        t.printStackTrace(ps);
        ps.close();
        
        return new String(bos.toByteArray());
    }
    
    private void append(final String s, final Color color) {
        if (s == null) {
            return;
        }
        Display.getDefault().asyncExec(
            new Runnable() {
                public void run() {
                    try {
                        StyleRange styleRange = new StyleRange();
                        styleRange.start = document.getLength() - 1;
                        styleRange.length = s.length() + nl.length();
                        styleRange.foreground = color;
                        
                        document.replace(document.getLength(), 0, s);
                        document.replace(document.getLength(), 0, nl);

                        if (console.getTextWidget() != null) { // amazingly, this happens...
                            console.getTextWidget().setStyleRange(styleRange);
                        }
                    } catch (BadLocationException e) {
                        e.printStackTrace();
                    }
                }
            }
        );
    }

    /**
     * @inheritdoc
     */
    public void error(String arg0) {
        append(arg0, colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
    }

    /**
     * @inheritdoc
     */
    public void error(String arg0, Throwable arg1) {
        append(arg0, colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
        append(toString(arg1), colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
        
    }

    /**
     * @inheritdoc
     */
    public void fatalError(String arg0) {
        append(arg0, colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
    }

    /**
     * @inheritdoc
     */
    public void fatalError(String arg0, Throwable arg1) {
        append(arg0, colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
        append(toString(arg1), colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM));
    }
    
    public class LogHandler extends Handler {

        @Override
        public void publish(LogRecord record) {
            Color color = null;
            if (record.getLevel().intValue() <= Level.INFO.intValue()) {
                color = colorProvider.getColor(IDebugUIConstants.ID_STANDARD_OUTPUT_STREAM);
            } else {
                color = colorProvider.getColor(IDebugUIConstants.ID_STANDARD_ERROR_STREAM);
            } 
            append(record.getMessage(), color);
            if (record.getThrown() != null) {
                append(RenderProgress.this.toString(record.getThrown()), color);
            }
        }

        @Override
        public void flush() {
            // TODO Auto-generated method stub
            
        }

        @Override
        public void close() throws SecurityException {
            // TODO Auto-generated method stub
            
        }
        
    }
    
    // Progress Monitor -------------------------------------------------------
    
    public class ProgressMonitor implements IProgressMonitor {

        private String name;
        private boolean canceled;

        /**
         * @inheritdoc
         */
        public void beginTask(final String name, final int totalWork) {
            Display.getDefault().asyncExec(
                new Runnable() {
                    public void run() {
                        progressBar.setMinimum(0);
                        progressBar.setSelection(0);
                        progressBar.setMaximum(totalWork);

                        progressCaption.setText("Progress: " + name);
                        ProgressMonitor.this.name = name;
                    }
                }
            );
        }

        /**
         * @inheritdoc
         */
        public void done() {
            Display.getDefault().asyncExec(
                new Runnable() {
                    public void run() {
                        progressBar.setSelection(progressBar.getMaximum());
                        progressCaption.setText("Progress: Done");
                    }
                }
            );
        }

        /**
         * @inheritdoc
         */
        public void internalWorked(double work) {
            worked((int) Math.round(work));
        }

        /**
         * @inheritdoc
         */
        public boolean isCanceled() {
            return canceled;
        }

        /**
         * @inheritdoc
         */
        public void setCanceled(boolean value) {
            this.canceled = value;
            if (value) {
                Display.getDefault().asyncExec(
                    new Runnable() {
                        public void run() {
                            progressCaption.setText("Progress: " + ProgressMonitor.this.name + "... Requesting cancelation");
                        }
                    }
                );
            }
        }

        /**
         * @inheritdoc
         */
        public void setTaskName(String name) {
            this.name = name;
        }

        /**
         * @inheritdoc
         */
        public void subTask(final String name) {
            Display.getDefault().asyncExec(
                new Runnable() {
                    public void run() {
                        progressCaption.setText("Progress: " + ProgressMonitor.this.name + "... " + name);
                    }
                }
            );
        }

        /**
         * @inheritdoc
         */
        public void worked(final int work) {
            Display.getDefault().asyncExec(
                new Runnable() {
                    public void run() {
                        progressBar.setSelection(progressBar.getSelection() + work);
                    }
                }
            );
        }
        
    }
}
